@@ -14,10 +14,15 @@
 #    You should have received a copy of the GNU General Public License
 #    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
 
+import cairo
 import math
 import logging
 
-from paperwork.backend.util import image2surface
+from gi.repository import Gtk
+from gi.repository import Pango
+from gi.repository import PangoCairo
+
+from paperwork_backend.util import image2surface
 
 
 logger = logging.getLogger(__name__)
@@ -29,7 +34,9 @@ class Drawer(object):
     IMG_LAYER = 200
     BOX_LAYER = 50
     PROGRESSION_INDICATOR_LAYER = 25
-    FADDING_EFFECT_LAYER = 0
+    BUTTON_LAYER = 20
+    FADDING_EFFECT_LAYER = 10
+    CURSOR_LAYER = 0
     # layer number == priority --> lower is drawn last (higher level)
 
     layer = -1  # must be set by subclass
@@ -57,7 +64,10 @@ class Drawer(object):
             should_be_visible = False
         return should_be_visible
 
-    def draw_surface(self, cairo_ctx, surface, img_position, img_size, angle=0):
+    def draw_surface(self,
+                     cairo_ctx, surface, img_position, img_size, angle=0,
+                     absolute=False,
+                     ):
         """
         Draw a surface
 
@@ -81,11 +91,16 @@ class Drawer(object):
         cairo_ctx.save()
         try:
             cairo_ctx.translate(img_position[0], img_position[1])
-            cairo_ctx.translate(-self.canvas.offset[0], -self.canvas.offset[1])
+            if absolute:
+                cairo_ctx.translate(
+                    self.canvas.offset[0], self.canvas.offset[1]
+                )
             if angle != 0:
                 cairo_ctx.translate(img_size[0] / 2, img_size[1] / 2)
                 cairo_ctx.rotate(angle)
                 cairo_ctx.translate(-img_size[0] / 2, -img_size[1] / 2)
+            if scaling[0] <= 0.0 or scaling[1] <= 0.0:
+                return
             cairo_ctx.scale(scaling[0], scaling[1])
 
             cairo_ctx.set_source_surface(
@@ -99,12 +114,6 @@ class Drawer(object):
             cairo_ctx.restore()
 
     def do_draw(self, cairo_ctx):
-        """
-        Arguments:
-            offset --- Position of the area in which to draw:
-                       (offset_x, offset_y)
-            size --- Size of the area in which to draw: (width, height) = size
-        """
         assert()
 
     def on_tick(self):
@@ -137,12 +146,7 @@ class Drawer(object):
                 position = (position[0] - (diff / 2), position[1])
             else:
                 position = (position[0], position[1] - (diff / 2))
-
-        p = (max(0, position[0] - self.canvas.offset[0]),
-             max(0, position[1] - self.canvas.offset[1]))
-        p = (min(p[0], self.canvas.size[0]),
-             min(p[1], self.canvas.size[1]))
-        return p
+        return position
 
     relative_position = property(_get_relative_position)
 
@@ -162,10 +166,6 @@ class Drawer(object):
 
         edge = (position[0] + size[0],
                 position[1] + size[1])
-        edge = (max(0, edge[0] - self.canvas.offset[0]),
-                max(0, edge[1] - self.canvas.offset[1]))
-        edge = (min(edge[0], self.canvas.size[0]),
-                min(edge[1], self.canvas.size[1]))
         return edge
 
     def _get_relative_size(self):
@@ -176,9 +176,27 @@ class Drawer(object):
 
     relative_size = property(_get_relative_size)
 
-    def redraw(self):
-        self.canvas.redraw((self.relative_position,
-                            self.relative_size))
+    def redraw(self, extra_border=0):
+        if not self.visible:
+            logger.warning("redraw() called on non-visible element")
+            return
+
+        position = self.relative_position
+        size = self.relative_size
+
+        position = (
+            max(0, position[0] - extra_border),
+            max(0, position[1] - extra_border)
+        )
+        size = (
+            size[0] + (2 * extra_border),
+            size[1] + (2 * extra_border)
+        )
+
+        self.canvas.redraw((
+            position,
+            size
+        ))
 
     def show(self):
         pass
@@ -187,6 +205,56 @@ class Drawer(object):
         pass
 
 
+class Centerer(Drawer):
+    """
+    Wrapper to keep a drawer centered
+    """
+    def __init__(self, child):
+        super(Centerer, self).__init__()
+        self.child = child
+
+    def set_canvas(self, canvas):
+        super(Centerer, self).set_canvas(canvas)
+        self.child.set_canvas(canvas)
+
+    def __get_layer(self):
+        return self.child.layer
+
+    def __set_layer(self, v):
+        self.child.layer = v
+
+    layer = property(__get_layer, __set_layer)
+
+    def __get_size(self):
+        return self.child.size
+
+    size = property(__get_size)
+
+    def __get_position(self):
+        if not self.canvas:
+            logger.warning(
+                "Centerer: Canvas not set yet. Cannot center drawer !"
+            )
+            return self.child.position
+        return (
+            self.child.position[0] + (self.canvas.visible_size[0] / 2),
+            self.child.position[1] + (self.canvas.visible_size[1] / 2),
+        )
+
+    position = property(__get_position)
+
+    def do_draw(self, cairo_ctx):
+        cairo_ctx.save()
+        try:
+            cairo_ctx.translate(
+                int(self.canvas.visible_size[0] / 2),
+                int(self.canvas.visible_size[1] / 2)
+            )
+            self.child.do_draw(cairo_ctx)
+        finally:
+            cairo_ctx.restore()
+
+
 class BackgroundDrawer(Drawer):
     layer = Drawer.BACKGROUND_LAYER
 
@@ -208,6 +276,81 @@ class BackgroundDrawer(Drawer):
         cairo_ctx.paint()
 
 
+class CursorDrawer(Drawer):
+    """
+    Beware ! This drawer is *not* relative.
+    """
+    layer = Drawer.CURSOR_LAYER
+
+    def __init__(self, gdk_cursor, position):
+        self.canvas = None
+        self._position = position
+
+        img = gdk_cursor.get_image()
+        self.size = (img.get_width(), img.get_height())
+
+        (self.cursor_surface, hotspot_x, hotspot_y) = gdk_cursor.get_surface()
+        self.hotspot = (hotspot_x, hotspot_y)
+        self.visible = True
+
+    def set_canvas(self, canvas):
+        self.canvas = canvas
+
+    @staticmethod
+    def compute_visibility(offset, visible_area_size, position, size):
+        return True
+
+    def draw(self, cairo_ctx):
+        if not self.visible:
+            return
+        self.draw_surface(
+            cairo_ctx, self.cursor_surface, self.position, self.size,
+            absolute=True
+        )
+
+    def _get_position(self):
+        return (
+            self._position[0] - self.hotspot[0],
+            self._position[1] - self.hotspot[1]
+        )
+
+    position = property(_get_position)
+
+    def _get_relative_position(self):
+        return self.position
+
+    relative_position = property(_get_relative_position)
+
+    def _get_relative_size(self):
+        return self.size
+
+    relative_size = property(_get_relative_size)
+
+    def redraw(self, extra_border=0):
+        position = self.position
+        size = self.size
+
+        position = (
+            max(0, position[0] - extra_border),
+            max(0, position[1] - extra_border)
+        )
+        size = (
+            size[0] + (2 * extra_border),
+            size[1] + (2 * extra_border)
+        )
+
+        self.canvas.redraw((
+            position,
+            size
+        ))
+
+    def show(self):
+        self.visible = True
+
+    def hide(self):
+        self.visible = False
+
+
 class RectangleDrawer(Drawer):
     layer = Drawer.BOX_LAYER
     visible = True
@@ -239,19 +382,19 @@ class RectangleDrawer(Drawer):
 
             if self.angle != 0:
                 angle = math.pi * self.angle / 180
-                cairo_ctx.translate(self.position[0] - self.canvas.offset[0]
-                                    + (self.size[0] / 2),
-                                    self.position[1] - self.canvas.offset[1]
-                                    + (self.size[1] / 2))
+                cairo_ctx.translate(self.position[0] +
+                                    (self.size[0] / 2),
+                                    self.position[1] +
+                                    (self.size[1] / 2))
                 cairo_ctx.rotate(angle)
-                cairo_ctx.translate(-self.position[0] + self.canvas.offset[0]
-                                    - (self.size[0] / 2),
-                                    -self.position[1] + self.canvas.offset[1]
-                                    - (self.size[1] / 2))
+                cairo_ctx.translate(-self.position[0] -
+                                    (self.size[0] / 2),
+                                    -self.position[1] -
+                                    (self.size[1] / 2))
 
             cairo_ctx.rectangle(
-                self.position[0] - self.canvas.offset[0],
-                self.position[1] - self.canvas.offset[1],
+                self.position[0],
+                self.position[1],
                 self.size[0], self.size[1]
             )
             cairo_ctx.clip()
@@ -312,10 +455,10 @@ class LineDrawer(Drawer):
             cairo_ctx.set_source_rgba(self.color[0], self.color[1],
                                       self.color[2], self.color[3])
             cairo_ctx.set_line_width(self.width)
-            cairo_ctx.move_to(self.start[0] - self.canvas.offset[0],
-                              self.start[1] - self.canvas.offset[1])
-            cairo_ctx.line_to(self.end[0] - self.canvas.offset[0],
-                              self.end[1] - self.canvas.offset[1])
+            cairo_ctx.move_to(self.start[0],
+                              self.start[1])
+            cairo_ctx.line_to(self.end[0],
+                              self.end[1])
             cairo_ctx.stroke()
         finally:
             cairo_ctx.restore()
@@ -326,7 +469,7 @@ class PillowImageDrawer(Drawer):
     visible = True
 
     def __init__(self, position, image):
-        Drawer.__init__(self)
+        super(PillowImageDrawer, self).__init__()
         self.size = image.size
         self.img_size = self.size
         self.position = position
@@ -339,6 +482,58 @@ class PillowImageDrawer(Drawer):
                           self.size, self.angle)
 
 
+class TextDrawer(Drawer):
+    """
+    Centered text
+    """
+    layer = Drawer.IMG_LAYER
+    visible = True
+
+    def __init__(self, position, text, height=12):
+        super(TextDrawer, self).__init__()
+        self.size = (10, 10)  # will be updated later when text will be rendered
+        self.center_position = position
+        # In this case, it's actually the center of the text.
+        # Will be updated later to be the top-left
+        self.position = position
+        self.angle = 0
+        self.text = text
+        self.height = height
+        self.font = None
+
+    def do_draw(self, cairo_ctx):
+        cairo_ctx.save()
+        try:
+            layout = PangoCairo.create_layout(cairo_ctx)
+            if self.font:
+                font = Pango.FontDescription()
+                font.set_family(self.font)
+                layout.set_font_description(font)
+            layout.set_text(self.text, -1)
+
+            txt_size = layout.get_size()
+            if 0 in txt_size:
+                return
+            txt_factor = float(self.height) / txt_size[1]
+            self.size = (
+                int(txt_size[0] * txt_factor),
+                int(txt_size[1] * txt_factor)
+            )
+            self.position = (
+                int(self.center_position[0] - (self.size[0] / 2)),
+                int(self.center_position[1] - (self.size[1] / 2))
+            )
+            cairo_ctx.translate(self.position[0], self.position[1])
+            if txt_factor <= 0.0:
+                return
+            cairo_ctx.scale(txt_factor * Pango.SCALE, txt_factor * Pango.SCALE)
+            cairo_ctx.set_source_rgb(0.0, 0.0, 0.0)
+            PangoCairo.update_layout(cairo_ctx, layout)
+            PangoCairo.show_layout(cairo_ctx, layout)
+        finally:
+            cairo_ctx.restore()
+
+
 class TargetAreaDrawer(Drawer):
     layer = Drawer.BOX_LAYER
     visible = True
@@ -475,12 +670,12 @@ class TargetAreaDrawer(Drawer):
                 func,
                 (
                     (
-                        rect[0][0] - self.canvas.offset[0],
-                        rect[0][1] - self.canvas.offset[1],
+                        rect[0][0],
+                        rect[0][1],
                     ),
                     (
-                        rect[1][0] - self.canvas.offset[0],
-                        rect[1][1] - self.canvas.offset[1],
+                        rect[1][0],
+                        rect[1][1],
                     ),
                 )
             )
@@ -491,6 +686,116 @@ class TargetAreaDrawer(Drawer):
             func(cairo_ctx, rect)
 
 
+class ProgressBarDrawer(Drawer):
+    layer = Drawer.PROGRESSION_INDICATOR_LAYER
+    visible = True
+
+    TXT_MARGIN = 5
+
+    def __init__(self,
+                 val_min=0, val_max=100, val_current=0, text=u"",
+                 back_color=(0.75, 0.75, 0.75),
+                 bar_color=(0.6, 0.6, 0.6),
+                 text_color=(0.0, 0.0, 0.0)):
+        Drawer.__init__(self)
+        self.val_min = val_min
+        self.val_max = val_max
+        self.val_current = val_current
+        self.back_color = back_color
+        self.bar_color = bar_color
+        self.text_color = text_color
+        self.text = text
+        self.__last_text_height = 50
+
+    def set_progression(self, val_current, text, val_min=None, val_max=None):
+        self.val_current = val_current
+        self.text = text
+        if val_min is not None:
+            self.val_min = val_min
+        if val_max is not None:
+            self.val_max = val_max
+        self.redraw()
+
+    def redraw(self):
+        txt_h = self.__last_text_height
+        position = (0, self.canvas.size[1] - (txt_h + (2 * self.TXT_MARGIN)))
+        size = (self.canvas.size[0] + (2 * self.TXT_MARGIN),
+                (txt_h + 2 * self.TXT_MARGIN))
+        self.canvas.redraw((position, size))
+
+    def _get_position(self):
+        if not self.canvas:
+            return (0, 0)
+        offset = self.canvas.offset
+        size = self.canvas.size
+        return (offset[0] + 1,
+                size[1] + offset[1] -
+                (self.__last_text_height + (2 * self.TXT_MARGIN)) - 1)
+
+    position = property(_get_position)
+
+    def _get_size(self):
+        if not self.canvas:
+            return (200, self.__last_text_height + 2 * self.TXT_MARGIN)
+        return (self.canvas.size[0] - 1,
+                (self.__last_text_height + 2 * self.TXT_MARGIN))
+
+    size = property(_get_size)
+
+    def draw(self, cairo_ctx):
+        if not self.visible:
+            return
+        txt = self.text
+        if txt is None or txt == "":
+            txt = u"T"
+
+        cairo_ctx.select_font_face("", cairo.FONT_SLANT_NORMAL,
+                                   cairo.FONT_WEIGHT_BOLD)
+        txt_ext = cairo_ctx.text_extents(txt)
+        (p_x1, p_y1, p_x2, p_y2, p_x3, p_y3) = txt_ext
+        txt_h = p_y2
+        self.__last_text_height = txt_h
+
+        position = self.position
+        size = self.size
+
+        (canvas_w, canvas_h) = self.canvas.size
+
+        w = ((canvas_w * (self.val_current - self.val_min)) /
+             (self.val_max - self.val_min))
+
+        cairo_ctx.save()
+        try:
+            for (color, rect_pos, rect_size) in [
+                (
+                    self.bar_color,
+                    (position[0], position[1]),
+                    (w, size[1])
+                ),
+                (
+                    self.back_color,
+                    (position[0] + w, position[1]),
+                    (canvas_w - position[0] - w, size[1])
+                ),
+            ]:
+                cairo_ctx.set_source_rgb(color[0], color[1], color[2])
+                cairo_ctx.rectangle(
+                    rect_pos[0], rect_pos[1],
+                    rect_size[0], rect_size[1]
+                )
+                cairo_ctx.fill()
+        finally:
+            cairo_ctx.restore()
+
+        cairo_ctx.set_source_rgb(self.text_color[0], self.text_color[1],
+                                 self.text_color[2])
+        cairo_ctx.move_to(self.TXT_MARGIN,
+                          self.canvas.size[1] + self.canvas.offset[1] -
+                          self.TXT_MARGIN)
+        cairo_ctx.text_path(self.text)
+        cairo_ctx.fill()
+
+
 def fit(element_size, area_size, force=False):
     """
     Return the size to give to the element so it fits in the area size.
